@@ -2,7 +2,7 @@ import os
 
 __author__ = """Camille Scott"""
 __email__ = 'cswel@ucdavis.edu'
-__version__ = '0.9.4'
+__version__ = '1.0.0'
 
 __splash__ = f'''
  _____             _      _____
@@ -15,4 +15,6 @@ __splash__ = f'''
 __about__ = 'A discord dice-rolling bot with lots of nifty features.'
 __testing__ = bool(os.environ.get('ZARDOZ_DEBUG', False))
 
-from .zardoz import main
+from .cli import main
+from . import cogs
+from . import dice
